#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <string.h>


#include "asap.h"

#include "myAll.h"
#include "agatGen.h"
#include "utils.h"
#include "agatsend.h"

#include "libAGAT.h"



/******************/
/* structures ... */
/******************/


/* cells in the hash-lists */
typedef struct FidName {
    Fid             fid;
    char           *name;
    struct FidName *next;
}               FidName;



/**************/
/* local vars */
/**************/

/* pour accepter la connection */
static char    *socketName;


/* name and FILE* of the dump file to use */
static char    *dumpFileName = (char *) NULL;
static FILE    *dumpFile;

/* vars used for communication */
static ASAPserv_t serv;
static ASAPconn_t conn;
static channel_t chan;
static ASAPstatus_t err;

/* has the initialisation been done? */
static Boolean  initDone = False;

/* what's to do? Create a dump-file or (and) play an anim */
static Boolean  dump = False;
static Boolean  play = False;
static Boolean  run = False;

/* for prebuild msg header storage */
static MsgHeader *tMsgHeader;

/* hash table for stream-name-storage (to retrive Fid from stream-name) */
static FidName *tHashFidName[HASHT_SZ];


/* to find user and system time used between two send */
static long     lastTime = 0;



/**************/
/* prototypes */
/**************/

static void     init(void);
static void     getEnvParams(char **psn, char **pdfn);
static void     acceptConnect(char *sn);
static void     openDumpFile(char *dumpFileName);
static void     checkAsapStat(ASAPstatus_t stat, char *errMsg);
static Fid      nameToFid(char *name, int stype, int vtype);
static Fid      searchListName(char *name, FidName * list);
static Fid      addName(char *name, int stype, int vtype);
static int      hashName(char *name);
static long     dTime(void);


/**********/
/* bodies */
/**********/



/* fonctions necessaires aux communications */

static void     init(void)
{
    getEnvParams(&socketName, &dumpFileName);
    if (run != True)
	return;
    if (play == True)
	acceptConnect(socketName);
    if (dump == True)
	openDumpFile(dumpFileName);
    initDone = True;
}



static void     getEnvParams(char **psn, char **pdfn)
{
    char            tmp[MAX_CHAR_TMP];
    char           *s;

    sprintf(tmp, "%s%s", AGAT_ENV, "socket");
    socketName = getenv(tmp);


    sprintf(tmp, "%s%s", AGAT_ENV, "play");
    s = getenv(tmp);
    if ((s != NULL) && (strlen(s) != 0)) {
	play = True;
    }

    sprintf(tmp, "%s%s", AGAT_ENV, "dump");
    *pdfn = getenv(tmp);
    if ((*pdfn != NULL) && (strlen(*pdfn) != 0)) {
	dump = True;
    }
    sprintf(tmp, "%s%s", AGAT_ENV, "running");
    s = getenv(tmp);
    if ((s != NULL) && (strlen(s) != 0)) {
	run = True;
    }
}


static void     acceptConnect(char *sn)
{
    char            socketname[1000];

    ASAPlocalConnect(sn, &conn);
    chan = conn.normal;
    initDone = True;
}




static void     openDumpFile(char *dumpFileName)
{
    struct rusage   rusage;

    dump = True;
    dumpFile = UCompressedFOpen(dumpFileName, "w");
    if (dumpFile == NULL) {
	error2(ER_CANTOPENDF, dumpFileName);
    }
    getrusage(RUSAGE_SELF, &rusage);
    lastTime = rusage.ru_utime.tv_sec * 1000 + rusage.ru_utime.tv_usec * 1e-3 +
	rusage.ru_stime.tv_sec * 1000 + rusage.ru_stime.tv_usec * 1e-3;
}


static void     checkAsapStat(ASAPstatus_t stat, char *errMsg)
{
    if ((stat != ASAP_SUCCESS) && (stat != ASAP_FINISH))
	error3("lib : ", errMsg, ASAPerrorText(stat));
}

static Fid      nameToFid(char *name, int stype, int vtype)
{
    static char    *lastName = (char *) NULL;
    static int      lastFid = 0;

    if ((lastName != (char *) NULL) && (strcmp(lastName, name) == 0)) {
	return lastFid;
    }
    else {
	if (lastName != (char *) NULL)
	    free(lastName);
	lastName = myStrDup(name);
	lastFid = searchListName(name, tHashFidName[hashName(name)]);
	if (lastFid == 0)
	    lastFid = addName(name, stype, vtype);
	return lastFid;
    }
}



static Fid      searchListName(char *name, FidName * list)
{
    while (list != (FidName *) NULL) {
	if (strcmp(list->name, name) == 0)
	    return list->fid;
	else
	    list = list->next;
    }
    return 0;
}

static Fid      addName(char *name, int stype, int vtype)
{
    static          tMsgHeaderSz = 0;
    static          nbMsgHeader = 0;
    SDescStream     sdf;

    int             hind;	/* hash index */
    FidName        *nc;		/* new cell */

    /* create and fill a new cell in the Hash list */
    hind = hashName(name);
    nc = (FidName *) malloc(sizeof(FidName));
    nc->next = tHashFidName[hind];
    nc->name = myStrDup(name);

    /* create a new fid */
    if (nbMsgHeader >= tMsgHeaderSz) {
	if (tMsgHeader == (MsgHeader *) NULL) {
	    tMsgHeaderSz += RM_CHKZ_MSGHEADER;
	    tMsgHeader = (MsgHeader *) malloc(sizeof(MsgHeader) * tMsgHeaderSz);
	}
	else {
	    tMsgHeaderSz += RM_CHKZ_MSGHEADER;
	    tMsgHeader = (MsgHeader *) realloc(tMsgHeader, sizeof(MsgHeader) * tMsgHeaderSz);
	}
    }
    nbMsgHeader++;		/* to avoid using fid ==0 */
    nc->fid = nbMsgHeader;
    tHashFidName[hind] = nc;

    /* fill a new prebuild msg header cell */
    tMsgHeader[nbMsgHeader].type = MELT_SVTYPE(stype, vtype);
    tMsgHeader[nbMsgHeader].fid = 0;

    /* send the new fid description to the server */
    err = ASAPsendBinary(chan, (char *) &tMsgHeader[nbMsgHeader], sizeof(MsgHeader));
    checkAsapStat(err, "send new fid");
    sdf.fid = nbMsgHeader;
    sdf.nameLen = strlen(name) + 1;
    err = ASAPsendBinary(chan, (char *) &sdf, sizeof(SDescStream));
    checkAsapStat(err, "send new fid");
    err = ASAPsendBinary(chan, name, sdf.nameLen);
    checkAsapStat(err, "send new fid");

    /* finish to fill prebuild msg header cell */
    tMsgHeader[nbMsgHeader].fid = nbMsgHeader;

    return nbMsgHeader;
}




static int      hashName(char *name)
{
    return name[0] % HASHT_SZ;
}




static long     dTime(void)
{
    long            dif;
    long            newTime;
    struct rusage   newRusage;

    getrusage(RUSAGE_SELF, &newRusage);
    newTime = newRusage.ru_utime.tv_sec * 1000 + newRusage.ru_utime.tv_usec * 1e-3 +
	newRusage.ru_stime.tv_sec * 1000 + newRusage.ru_stime.tv_usec * 1e-3;
    dif = newTime - lastTime;
    lastTime = newTime;
    if (dif == 0)
	return 1;
    return dif;
}


/**********************/
/* exported functions */
/**********************/



void            agatSendChar(char *f, char v)
{
    Fid             fid;

    if (initDone == False)
	init();
    if (run != True)
	return;
    if (dump == True) {
	fprintf(dumpFile, "%s %d %c %d\n", f, MELT_SVTYPE(STYP_SVAL, VTYP_CHAR), v, dTime());
    }
    if (play == True) {
	fid = nameToFid(f, STYP_SVAL, VTYP_CHAR);
	err = ASAPsendBinary(chan, (char *) &tMsgHeader[fid], sizeof(MsgHeader));
	checkAsapStat(err, "send char header");
	err = ASAPsendBinary(chan, (char *) &v, sizeof(char));
	checkAsapStat(err, "send char");
	ASAPflush(chan);
    }
}

void            agatSendUChar(char *f, unsigned char v)
{
    Fid             fid;

    if (initDone == False)
	init();
    if (run != True)
	return;
    if (dump == True) {
	fprintf(dumpFile, "%s %d %c %d\n", f, MELT_SVTYPE(STYP_SVAL, VTYP_UCHAR), v, dTime());
    }
    if (play == True) {
	fid = nameToFid(f, STYP_SVAL, VTYP_UCHAR);
	err = ASAPsendBinary(chan, (char *) &tMsgHeader[fid], sizeof(MsgHeader));
	checkAsapStat(err, "send unsigned char header");
	err = ASAPsendBinary(chan, (char *) &v, sizeof(unsigned char));
	checkAsapStat(err, "send unsigned char");
	ASAPflush(chan);
    }
}


void            agatSendShort(char *f, short int v)
{
    Fid             fid;


    if (initDone == False)
	init();
    if (run != True)
	return;
    if (dump == True) {
	fprintf(dumpFile, "%s %d %d %d\n", f, MELT_SVTYPE(STYP_SVAL, VTYP_SHORT), v, dTime());
    }
    if (play == True) {
	fid = nameToFid(f, STYP_SVAL, VTYP_SHORT);
	err = ASAPsendBinary(chan, (char *) &tMsgHeader[fid], sizeof(MsgHeader));
	checkAsapStat(err, "send short header");
	err = ASAPsendBinary(chan, (char *) &v, sizeof(short));
	checkAsapStat(err, "send short ");
	ASAPflush(chan);
    }
}

void            agatSendUShort(char *f, short unsigned int v)
{
    Fid             fid;


    if (initDone == False)
	init();
    if (run != True)
	return;
    if (dump == True) {
	fprintf(dumpFile, "%s %d %d %d\n", f, MELT_SVTYPE(STYP_SVAL, VTYP_USHORT), v, dTime());
    }
    if (play == True) {
	fid = nameToFid(f, STYP_SVAL, VTYP_USHORT);
	err = ASAPsendBinary(chan, (char *) &tMsgHeader[fid], sizeof(MsgHeader));
	checkAsapStat(err, "send unsigned short header");
	err = ASAPsendBinary(chan, (char *) &v, sizeof(unsigned short));
	checkAsapStat(err, "send unsigned short ");
	ASAPflush(chan);
    }
}


void            agatSendInt(char *f, int v)
{
    Fid             fid;


    if (initDone == False)
	init();
    if (run != True)
	return;
    if (dump == True) {
	fprintf(dumpFile, "%s %d %d %d\n", f, MELT_SVTYPE(STYP_SVAL, VTYP_INT), v, dTime());
    }
    if (play == True) {
	fid = nameToFid(f, STYP_SVAL, VTYP_INT);
	err = ASAPsendBinary(chan, (char *) &tMsgHeader[fid], sizeof(MsgHeader));
	checkAsapStat(err, "send int header");
	err = ASAPsendBinary(chan, (char *) &v, sizeof(int));
	checkAsapStat(err, "send int");
	ASAPflush(chan);
    }
}

void            agatSendUInt(char *f, unsigned int v)
{
    Fid             fid;


    if (initDone == False)
	init();
    if (run != True)
	return;
    if (dump == True) {
	fprintf(dumpFile, "%s %d %d %d\n", f, MELT_SVTYPE(STYP_SVAL, VTYP_UINT), v, dTime());
    }
    if (play == True) {
	fid = nameToFid(f, STYP_SVAL, VTYP_UINT);
	err = ASAPsendBinary(chan, (char *) &tMsgHeader[fid], sizeof(MsgHeader));
	checkAsapStat(err, "send unsigned int header");
	err = ASAPsendBinary(chan, (char *) &v, sizeof(unsigned int));
	checkAsapStat(err, "send unsigned int");
	ASAPflush(chan);
    }
}


void            agatSendLong(char *f, long int v)
{
    Fid             fid;


    if (initDone == False)
	init();
    if (run != True)
	return;
    if (dump == True) {
	fprintf(dumpFile, "%s %d %d %d\n", f, MELT_SVTYPE(STYP_SVAL, VTYP_LONG), v, dTime());
    }
    if (play == True) {
	fid = nameToFid(f, STYP_SVAL, VTYP_LONG);
	err = ASAPsendBinary(chan, (char *) &tMsgHeader[fid], sizeof(MsgHeader));
	checkAsapStat(err, "send long header");
	err = ASAPsendBinary(chan, (char *) &v, sizeof(long));
	checkAsapStat(err, "send long ");
	ASAPflush(chan);
    }
}

void            agatSendUlong(char *f, long unsigned int v)
{
    Fid             fid;


    if (initDone == False)
	init();
    if (run != True)
	return;
    if (dump == True) {
	fprintf(dumpFile, "%s %d %d %d\n", f, MELT_SVTYPE(STYP_SVAL, VTYP_ULONG), v, dTime());
    }
    if (play == True) {
	fid = nameToFid(f, STYP_SVAL, VTYP_ULONG);
	err = ASAPsendBinary(chan, (char *) &tMsgHeader[fid], sizeof(MsgHeader));
	checkAsapStat(err, "send unsigned long header");
	err = ASAPsendBinary(chan, (char *) &v, sizeof(unsigned long));
	checkAsapStat(err, "send unsigned long");
	ASAPflush(chan);
    }
}


void            agatSendFloat(char *f, float v)
{
    Fid             fid;


    if (initDone == False)
	init();
    if (run != True)
	return;
    if (dump == True) {
	fprintf(dumpFile, "%s %d %.15g %d\n", f, MELT_SVTYPE(STYP_SVAL, VTYP_FLOAT), v, dTime());
    }
    if (play == True) {
	fid = nameToFid(f, STYP_SVAL, VTYP_FLOAT);
	err = ASAPsendBinary(chan, (char *) &tMsgHeader[fid], sizeof(MsgHeader));
	checkAsapStat(err, "send float header");
	err = ASAPsendBinary(chan, (char *) &v, sizeof(float));
	checkAsapStat(err, "send float");
	ASAPflush(chan);
    }
}


void            agatSendDouble(char *f, double v)
{
    Fid             fid;


    if (initDone == False)
	init();
    if (run != True)
	return;
    if (dump == True) {
	fprintf(dumpFile, "%s %d %.15g %d\n", f, MELT_SVTYPE(STYP_SVAL, VTYP_DOUBLE), v, dTime());
    }
    if (play == True) {
	fid = nameToFid(f, STYP_SVAL, VTYP_DOUBLE);
	err = ASAPsendBinary(chan, (char *) &tMsgHeader[fid], sizeof(MsgHeader));
	checkAsapStat(err, "send double header");
	err = ASAPsendBinary(chan, (char *) &v, sizeof(double));
	checkAsapStat(err, "send double");
	ASAPflush(chan);
    }
}
