/*
 */

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <string.h>

#include "param.h"
#include "asap.h"

#define CHECKFD(fd1,fd2,mes) {if ((fd1<0)||(fd2<0)) {perror(mes);exit(2);}}


#define MAXNB 10000

/* Connection in the UNIX domain */
static ASAPstatus_t localConnect(server, conn)
    ASAPserv_t      server;
    ASAPconn_t     *conn;
{
    int             i, len;
    int             nsd, usd;

    char            command[MAXCMDLEN];
    char            basename[MAXPATHLEN];

    struct sockaddr_un nad, uad;

    nsd = socket(AF_UNIX, SOCK_STREAM, 0);	/* the "normal" channel */
    usd = socket(AF_UNIX, SOCK_STREAM, 0);	/* the "urgent" channel */

    CHECKFD(nsd, usd, "cannot get socket descriptor in ASAPconnect");

    /* to be sure everything is clean... */
    bzero((char *) &nad, sizeof(nad));
    bzero((char *) &uad, sizeof(uad));

    nad.sun_family = uad.sun_family = AF_UNIX;

    /* construct the basenames : ASAP_UNIX_PATH<uid> */
    sprintf(basename, "%s%d", ASAP_UNIX_PATH, getuid());
    strcpy(nad.sun_path, basename);
    strcpy(uad.sun_path, basename);

    len = strlen(basename);

    for (i = 0; i < MAXNB; i++) {
	sprintf(nad.sun_path + len, "_%dn", i);
	sprintf(uad.sun_path + len, "_%du", i);
	/* the +2 is still a mystery for me... */
	if (bind(nsd, &nad, strlen(nad.sun_path) + 2) == 0) {
	    if (bind(usd, &uad, strlen(uad.sun_path) + 2) == 0) {
		break;
	    }
	    else {
		unlink(nad.sun_path);
		nsd = socket(AF_UNIX, SOCK_STREAM, 0);
		CHECKFD(nsd, usd, "cannot get socket descriptor in localConnect");
	    }
	}
    }

    /*
     * We store the X of ASAP_UNIX_PATH<uid>_x... so that ASAPterminate can
     * retrieve the file.
     */
    /* But it's up to the client !!! */
    if (i < MAXNB) {
	sprintf(basename + len, "_%d", i);
    }
    else
	return (ASAP_ENOPORT);

    listen(nsd, 1);
    listen(usd, 1);

    /* patch for Agat... must be incorporated into new versions of Asap */
    {
	char            tmp[1000];

	sprintf(tmp, "%s%s=%s", "AGAT", "socket", basename);
	putenv(myStrDup(tmp));
    }

    {
	char          **argv;
	int             pid, sexec;

	argv = (char **) server.cmd;
	if ((pid = fork()) == 0) {
	    sexec = execvp(argv[0], argv);
	    error2("Can't execute ", argv[0]);
	}
	else if (pid == -1) {
	    error2("Can't fork to execute ", argv[0]);
	}
    }

    /*
     * and accept connections ... Note that we begin with the normal
     * channel... Could mark the socket as NONBLOCKING and loop ?
     */
    {
	int             ansd, ausd, dummy;
	struct sockaddr_un target;

	ansd = accept(nsd, &target, &dummy);
	ausd = accept(usd, &target, &dummy);
	CHECKFD(ansd, ausd, "error in accept");
	close(nsd);
	close(usd);

	initConn(conn, ansd, ausd, nad.sun_path);
    }
    return ASAP_SUCCESS;
}


#define MAXPORT (IPPORT_RESERVED + 400)
/* connection in the INTERNET domain */
static ASAPstatus_t remoteConnect(server, conn)
    ASAPserv_t      server;
    ASAPconn_t     *conn;
{
    int             i;
    int             nsd, usd;
    int             port;
    struct sockaddr_in nad, uad;
    char            command[1024];
    extern void     GetLocalHostAddress();

    nsd = socket(AF_INET, SOCK_STREAM, 0);
    usd = socket(AF_INET, SOCK_STREAM, 0);
    /*
     * This is necessary on some UNIX (ULTRIX) to allow an immediate
     * "rebinding" after a failed "bind". We choose to get a fresh socket
     * instead. if ((setsockopt(nsd, SOL_SOCKET, SO_REUSEADDR, NULL, 0) < 0)
     * || (setsockopt(usd, SOL_SOCKET, SO_REUSEADDR, NULL, 0) < 0)) {
     * perror("cannot set SO_REUSEADDR option in remoteConnect"); exit(2); }
     */
    CHECKFD(nsd, usd, "cannot get socket descriptor in ASAPconnect");

    /* to be sure everything is clean... */
    bzero((char *) &nad, sizeof(nad));
    bzero((char *) &uad, sizeof(uad));

    nad.sin_family = AF_INET;
    uad.sin_family = AF_INET;
    nad.sin_addr.s_addr = INADDR_ANY;
    uad.sin_addr.s_addr = INADDR_ANY;

    for (i = IPPORT_RESERVED + 1; i < MAXPORT; i += 2) {
	nad.sin_port = htons((unsigned short) i);
	uad.sin_port = htons((unsigned short) (i + 1));
	if (bind(nsd, &nad, sizeof nad) == 0) {
	    if (bind(usd, &uad, sizeof uad) == 0) {
		break;
	    }
	    else {
		/*
		 * when "bind" failed, we obtain a "fresh" socket to avoid
		 * strange problems...
		 */
		close(nsd);
		close(usd);
		nsd = socket(AF_INET, SOCK_STREAM, 0);
		usd = socket(AF_INET, SOCK_STREAM, 0);
		CHECKFD(nsd, usd, "cannot get socket descriptor in remoteConnect");
	    }
	}
	else {
	    close(nsd);
	    nsd = socket(AF_INET, SOCK_STREAM, 0);
	    CHECKFD(nsd, usd, "cannot get socket descriptor in remoteConnect");
	}
    }

    if (i < MAXPORT) {
	port = i;
    }
    else
	return ASAP_ENOPORT;

    listen(nsd, 1);
    listen(usd, 1);

    /* Now, we launch the service ... */
    {
	char            local[MAXHOSTNAMELEN];

	GetLocalHostAddress(local);

	sprintf(command, "rsh %s %s --port %d --machine %s &",
		server.machine, server.cmd, port, local);
    }
    system(command);

    {
	int             ansd, ausd, dummy;
	struct sockaddr_in target;

	ansd = accept(nsd, &target, &dummy);
	ausd = accept(usd, &target, &dummy);
	if ((ansd < 0) || (ausd < 0)) {
	    perror("error in accept");
	    exit(2);
	}
	close(nsd);
	close(usd);

	initConn(conn, ansd, ausd, (char *) 0);
    }
    return ASAP_SUCCESS;

}

static ASAPstatus_t localConnectServer(server, conn)
    ASAPserv_t      server;
    ASAPconn_t     *conn;
{
    fatal("not yet implemented : localConnectServer");
}

static ASAPstatus_t remoteConnectServer(server, conn)
    ASAPserv_t      server;
    ASAPconn_t     *conn;
{
    fatal("not yet implemented : remoteConnectServer");
}


/* Should define what ASAPconnect returns... */

ASAPstatus_t    ASAPconnect(server, conn)
    ASAPserv_t      server;
    ASAPconn_t     *conn;
{
    switch (server.type) {
    case SERV_LOCAL:
	return (localConnect(server, conn));
    case SERV_REMOTE:
	return (remoteConnect(server, conn));
    case SERV_LOCAL_SERVER:
	return (localConnectServer(server, conn));
    case SERV_REMOTE_SERVER:
	return (remoteConnectServer(server, conn));
    default:
	fatal("unknow server type in ASAPconnect");
	/* NOTREACHED */
    }
    /* NOTREACHED */
}
