#include "param.h"
#include "asap.h"

#define AT_END(chan) (chan->in.curr == chan->in.last)
/* NEXT should never access beyond chan.in.last ... */
#define NEXT(chan) (chan->in.buf[chan->in.curr])
/* how much characters usable on "chan" */
#define AVAILABLE(chan) (chan->in.last - chan->in.curr + 1)

#define EATSEP(chan) if (getnext(chan) != SEP) fatal("not a separator")
#define CLEARTOK(chan) (chan->in.tok = ASAP_UNKNOWN)

static int      readChannel(chan)
    channel_t       chan;
{
    int             cc;

    if ((cc = read(chan->in.fd, chan->in.buf, chan->in.size)) < 0) {
	/* !!! plus fin ! La connection coupe'e ? */
	perror("read failed in readChannel");
	exit(ASAP_ESYS);
    }
    chan->in.curr = 0;
    chan->in.last = cc - 1;	/* attention !!! */
    return (cc);
}


static char     lookahead(chan)
    channel_t       chan;
{
    int             cc;

    if (chan->in.curr > chan->in.last) {
	if ((cc = readChannel(chan)) == 0) {
	    return (ASAP_BROKEN);
	}
    }
    return (chan->in.buf[chan->in.curr]);
}

/*
 * static char lookahead(chan) channel_t chan; {   int cc;
 * 
 * if (chan->in.curr > chan->in.last) { cc = readChannel(chan); } if (cc == 0) {
 * return(ASAP_BROKEN); } else { return(chan->in.buf[chan->in.curr]); }
 * 
 * } */


static char     getnext(chan)
    channel_t       chan;
{

    if (chan->in.curr > chan->in.last) {
	readChannel(chan);
    }
    return (chan->in.buf[chan->in.curr++]);
}


#define ISDECDIGIT(c) ((c >= '0') && (c <= '9'))
#define ISHEXDIGIT(c) ((c >= 'a') && (c <= 'f'))

static int      decodeInt(buf, len)
    char           *buf;
    int             len;
{
    int             i, b, r = 0;

    /* check length ... */

    /* first check the case of one digit */
    if (len == 1) {
	if (ISDECDIGIT(buf[0])) {
	    return (buf[0] - '0');
	}
	else {
	    fatal("not an integer in getInt");
	}
    }

    /* is it decimal or hexadecimal ? */
    if (buf[1] == 'x') {
	/* hexa */
	for (i = 2, b = 1; i < len; i++, b *= 16) {
	    if (ISDECDIGIT(buf[i])) {
		r += (buf[i] - '0') * b;
	    }
	    else {
		if (ISHEXDIGIT(buf[i])) {
		    r += (buf[i] - 'a' + 10) * b;
		}
		else {
		    fatal("not a decimal digit in decodeInt");
		}
	    }
	}
    }
    else {
	/* decimal */
	for (i = 2, b = 1; i < len; i++, b *= 10) {
	    if (ISDECDIGIT(buf[i])) {
		r += (buf[i] - '0') * b;
	    }
	    else {
		fatal("not a decimal digit in decodeInt");
	    }
	}
    }
    return (r);
}


/* should "eat" SEP ... */
static int      getInt(chan)
    channel_t       chan;
{
    char            buf[MAXARLEN];
    char            c;
    int             i;

    for (i = 0; ((c = getnext(chan)) != SEP) && (i <= MAXARLEN); i++) {
	buf[i] = c;
    }
    if (i > MAXARLEN) {
	fatal("arity too long !");
    }
    return (decodeInt(buf, i));
}


ASAPtoken_t     ASAPgetTokenType(chan)
    channel_t       chan;
{
    char            c;

    /* already known ? */
    if (chan->in.tok != ASAP_UNKNOWN) {
	return (chan->in.tok);
    }

    switch (c = lookahead(chan)) {
    case SEP:
	fatal("separator found by ASAPgetTokenType");
	break;			/* NOTREACHED */
    case BINTAG:
	chan->in.tok = ASAP_BIN;
	break;
    case ATTRTAG:
	chan->in.tok = ASAP_ATTROP;
	break;
    case ASAP_BROKEN:
	chan->in.tok = ASAP_BROKEN;
	break;
    default:
	if ((c >= '0') && (c <= '9')) {
	    chan->in.tok = ASAP_INT;
	}
	else {
	    chan->in.tok = ASAP_OP;
	}
    }
    return (chan->in.tok);
}

/*
 * Warning : "op" should be big enough (MAXOPLEN + 1). This is not
 * enforced...
 */
ASAPstatus_t    ASAPgetOp(chan, op, arity)
    channel_t       chan;
    char           *op;
    int            *arity;
{
    int             i;
    char            c;

    if (chan->in.tok == ASAP_UNKNOWN)
	ASAPgetTokenType(chan);
    if (chan->in.tok != ASAP_OP) {
	return ASAP_EBADTYPE;
    }

    /* get operator */
    for (i = 0; (c = getnext(chan)) != SEP; i++) {
	op[i] = c;
    }
    op[i] = '\0';		/* end of C string */

    /* get arity. note that SEP is "eaten"... */
    *arity = getInt(chan);

    CLEARTOK(chan);
    return ASAP_SUCCESS;
}

/*
 * Warning : "attr" should be big enough (MAXOPLEN + 1).
 */
ASAPstatus_t    ASAPgetAttrName(chan, attr)
    channel_t       chan;
    char           *attr;
{
    int             i;
    char            c;

    if (chan->in.tok != ASAP_UNKNOWN)
	return ASAP_EBADTYPE;

    for (i = 0; (c = getnext(chan)) != SEP; i++) {
	attr[i] = c;
    }
    attr[i] = '\0';		/* end of C string */
    CLEARTOK(chan);
    return ASAP_SUCCESS;
}

ASAPstatus_t    ASAPgetAttrOp(chan, op, arity, attrs)
    channel_t       chan;
    char           *op;
    int            *arity, *attrs;
{
    int             i;
    char            c;

    if (chan->in.tok == ASAP_UNKNOWN)
	ASAPgetTokenType(chan);
    if (chan->in.tok != ASAP_ATTROP) {
	return ASAP_EBADTYPE;
    }

    /* eat ATTRTAG */
    c = getnext(chan);

    /* get operator */
    for (i = 0; (c = getnext(chan)) != SEP; i++) {
	op[i] = c;
    }
    op[i] = '\0';		/* end of C string */

    /* get number of attributes "attrs" */
    *attrs = getInt(chan);

    /* get arity. note that SEP is "eaten"... */
    *arity = getInt(chan);

    CLEARTOK(chan);
    return ASAP_SUCCESS;
}


ASAPstatus_t    ASAPgetBinary(chan, buf, len, read)
    channel_t       chan;
    char           *buf;
    int             len;
    int            *read;
{
    int             i;
    int             total;	/* number of bytes in the binary object */

    if (chan->in.tok == ASAP_UNKNOWN)
	ASAPgetTokenType(chan);
    if (chan->in.tok != ASAP_BIN) {
	return ASAP_EBADTYPE;
    }

    /* eat "BINTAG" */
    if (getnext(chan) != BINTAG) {
	fatal("not a binary data in ASAPgetBinary");
    };

    /* get length */
    total = getInt(chan);

    /* while "buf" not full and "total" not reached */
    for (i = 0; (i < len) && (i < total); i++) {
	buf[i] = getnext(chan);
    }
    /* was "buf" big enough ? */
    if (i <= len) {
	EATSEP(chan);
	CLEARTOK(chan);
	*read = total;
	return (ASAP_FINISH);
    }
    else {
	chan->in.tok = ASAP_READING_BIN;
	chan->in.remaining = total - len;
	*read = len;
	return (ASAP_SUCCESS);
    }

}


ASAPstatus_t    ASAPgetMoreBinary(chan, buf, len, read)
    channel_t       chan;
    char           *buf;
    int             len;
    int            *read;
{
    int             i;

    if (chan->in.tok != ASAP_READING_BIN) {
	return ASAP_EBADTYPE;
    }
    /* while "buf" not full and "total" not reached */
    for (i = 0; (i < len) && (i < chan->in.remaining); i++) {
	buf[i] = getnext(chan);
    }
    /* was "buf" big enough ? */
    if (i <= len) {
	EATSEP(chan);
	CLEARTOK(chan);
	*read = chan->in.remaining;
	return (ASAP_FINISH);
    }
    else {
	chan->in.remaining -= len;
	*read = len;
	return (ASAP_SUCCESS);
    }

}

ASAPstatus_t    ASAPgetInt(chan, buf, len, read)
    channel_t       chan;
    char           *buf;
    int             len;
    int            *read;
{
    char            c;
    int             i;

    if (chan->in.tok == ASAP_UNKNOWN)
	ASAPgetTokenType(chan);
    if (chan->in.tok != ASAP_INT) {
	return ASAP_EBADTYPE;
    }


    /* while "buf" not full and "total" not reached */
    for (i = 0; (i < len) && ((c = getnext(chan)) != SEP); i++) {
	buf[i] = c;
    }
    /* was "buf" big enough ? */
    if (i <= len) {
	CLEARTOK(chan);
	*read = i;
	return (ASAP_FINISH);
    }
    else {
	chan->in.tok = ASAP_READING_INT;
	*read = len;
	return (ASAP_SUCCESS);
    }

}

ASAPstatus_t    ASAPgetMoreInt(chan, buf, len, read)
    channel_t       chan;
    char           *buf;
    int             len;
    int            *read;
{
    char            c;
    int             i;

    if (chan->in.tok != ASAP_READING_INT) {
	fatal("not reading");
    }
    /* while "buf" not full and "total" not reached */
    for (i = 0; (i < len) && ((c = getnext(chan)) != SEP); i++) {
	buf[i] = c;
    }
    /* was "buf" big enough ? */
    if (i <= len) {
	CLEARTOK(chan);
	*read = i;
	return (ASAP_FINISH);
    }
    else {
	*read = len;
	return (ASAP_SUCCESS);
    }

}
