#include <string.h>
#include <ctype.h>
/* for getpw... */
#include <pwd.h>
/* for MAXPATHLEN */
#include <sys/param.h>
/* for O_RDONLY */
#include <fcntl.h>

#include "param.h"

#include "asap.h"

extern char    *getenv();

/*
 * "GetHomeDir" returns the pathname of the user's home directory in "dest".
 * Returns a "void string" ("\0") if unable to find home-dir.
 */
static void     GetHomeDir(dest)
    char           *dest;
{

    uid_t           uid;
    extern struct passwd *getpwuid(), *getpwnam();

    struct passwd  *pw;
    register char  *ptr;

    if ((ptr = getenv("HOME")) != NULL) {
	(void) strcpy(dest, ptr);

    }
    else {
	if ((ptr = getenv("USER")) != NULL) {
	    pw = getpwnam(ptr);
	}
	else {
	    uid = getuid();
	    pw = getpwuid(uid);
	}
	if (pw) {
	    (void) strcpy(dest, pw->pw_dir);
	}
	else {
	    *dest = '\0';
	}
    }

}

/*
 * "fname" will be the pathname of the ASAP database of servers file. Note
 * that "fname" will contain DEFAULT_DBASE (i.e. "/.asapserv") if we are
 * unable to find home-dir.
 */

static void     GetFileName(fname)
    char           *fname;
{
    char           *servfile;

    if ((servfile = getenv(DEFAULT_DBASE_ENV_VAR)) == NULL) {
	/* we look in ~/.asapserv ... */
	fname[0] = '\0';
	GetHomeDir(fname);
	(void) strcat(fname, DEFAULT_DBASE);
	servfile = fname;
    }
    else
	strcpy(fname, servfile);
}

/*
 * The read buffer. For debugging, should be of a very low size to exercise
 * "looking_at" ...
 */
char            _buf[2048];

/* the (valid) length of _buf */
int             _buflen;

/* the current position (= next to be read) in _buf */
int             _current;

/* the file descriptor where to read from */
int             _fd;

#define INIT_IO(fd)  {_buflen = 0; _current = 1; _fd = fd;}


/* character for end-of-file... */
#define _EOF '\0'

#define _EOL -1
#define FAILED -1
#define EOF_RAISED 1
#define FOUND 20

/* the current character */
char            looking_at()
{
    if (_current < _buflen)
	return (_buf[_current]);
    else {
	if ((_buflen = read(_fd, _buf, 2048)) <= 0)
	    return (_EOF);
	else {
	    _current = 0;
	    return (_buf[0]);
	}
    }

}

/* ... */
#define SKIP() (_current++)

/* returns the current character and advances... */
char            next_char()
{
    char            c;

    c = looking_at();
    SKIP();
    return (c);
}


/*
 * machine:pgm arg1 ... argn machine!p 		: listening on port p and p +
 * 1 on "machine" !port			: listening on port p and p + 1 !path
 *  listening on socket path.normal and path.urgent
 */

static void     split_info(info, serv)
    char           *info;
    ASAPserv_t     *serv;
{
    int             len, inilen, machlen;

    len = strlen(info);
    inilen = strcspn(info, " \t");
    machlen = strcspn(info, ":");
    if ((machlen == len) || (machlen > inilen)) {
	serv->type = SERV_LOCAL;
	strcpy(serv->cmd, info);
    }
    else {
	serv->type = SERV_REMOTE;
	strncpy(serv->machine, info, machlen);
	strcpy(serv->cmd, info + (machlen + 1));
    }
}

/* advance till next non-white character and returns it */
static char     skip_white()
{
    char            c;

    while (((c = looking_at()) == ' ') || (c == '\t'))
	SKIP();
    return (c);
}

/* skip the remaining of the line */
static int      skip_line()
{
    char            c;

    while (((c = looking_at()) != '\n') || (c == _EOF))
	SKIP();
    if (c == _EOF)
	return (_EOF);
    else
	return (next_char());
}

#include <ctype.h>
#define TOKEN_CHAR(c) ((c > ' ') && (c <= '~'))
static int      get_token(tok)
    char           *tok;
{
    int             len = 0;

    while (TOKEN_CHAR(looking_at())) {
	tok[len] = next_char();
	len++;
    }
    return (len);
}

static int      get_rest(tok)
    char           *tok;
{
    int             len = 0;
    char            c;

    while (((c = looking_at()) != '\n') && (c != _EOF)) {
	tok[len] = next_char();
	len++;
    }
    return (len);
}


static int      get_line(servname, info)
    char           *servname;
    char           *info;
{
    char            tok[1024];
    int             len;

    switch (skip_white()) {
    case '#':
	skip_line();
	return (FAILED);
    case '\n':
	next_char();
	return (FAILED);
    case _EOF:
	return (EOF_RAISED);
    }

    if ((len = get_token(tok)) <= 0)
	return (EOF_RAISED);
    tok[len] = '\0';

    if (strcmp(tok, servname) != 0)
	return (FAILED);
    switch (skip_white()) {
    case '#':
	skip_line();
	return (FAILED);
    case '\n':
	next_char();
	return (FAILED);
    case _EOF:
	return (EOF_RAISED);
    }
    if ((len = get_rest(info)) <= 0)
	return (FAILED);
    info[len] = '\0';
    return (FOUND);


}


static int      findInFile(fname, servname, serv)
    char           *fname, *servname;
    ASAPserv_t     *serv;
{
    int             fd;
    int             status;

    char            info[1024];

    if ((fd = open(fname, O_RDONLY)) < 0)
	return FAILED;
    INIT_IO(fd);

    while ((status = get_line(servname, info)) == FAILED);
    close(fd);

    if (status == EOF_RAISED) {
	close(fd);
	return (FAILED);
    }

    if (status == FOUND) {	/* we've got it in "info"! */
	/* look at what is inside "info" */
	split_info(info, serv);
	return (FOUND);
    }
    else {
	close(fd);
	return FAILED;
    }

}

static ASAPstatus_t findInPath(path, servname, serv)
    char           *path, *servname;
    ASAPserv_t     *serv;
{
    char           *current;
    extern char    *strtok();

    current = strtok(path, ":");
    if (findInFile(current, servname, serv) == FOUND)
	return (ASAP_SUCCESS);

    while ((current = strtok(NULL, ":")) != NULL) {
	if (findInFile(current, servname, serv) == FOUND)
	    return (ASAP_SUCCESS);
    }
    return (ASAP_ENOSUCHSERV);
}


/* allocate the inside of a server structure */
void            initServ(serv)
    ASAPserv_t     *serv;
{
    char           *cmd, *machine, *socketname;

    if (((cmd = ASAPalloc(MAXCMDLEN + 1)) == NULL) ||
	((machine = ASAPalloc(MAXHOSTNAMELEN + 1)) == NULL) ||
	((socketname = ASAPalloc(MAXPATHLEN + 1)) == NULL)) {
	fatal("ASAPmalloc cannot allocate memory in ASAPallocServ");
    };
    serv->cmd = cmd;
    serv->machine = machine;
    serv->sockname = socketname;
}


ASAPstatus_t    ASAPgetservbyname(servname, serv)
    char           *servname;
    ASAPserv_t     *serv;
{
    char            fname[5 * MAXPATHLEN];

    /*
     * this is a potential memory leak as what is allocated for "serv" is
     * never freed...
     */
    initServ(serv);
    GetFileName(fname);
    return (findInPath(fname, servname, serv));

}
