/*
 * This file contains some useful fonctions.
 * 
 * 
 */


#include <sys/param.h>		/* for MAXHOSTNAMELEN */
#include <netdb.h>
#include <string.h>

#include "param.h"
#include "asap.h"

/*
 * "fatal" attempts to print "ASAP : msg\n" on the standard error file
 * descriptor and calls "exit" with 1. "msg" is truncated if longer than
 * "ERRBUFSIZE". To be used for irrecoverable errors. WARNING : we directly
 * write on file descriptor 2...
 */
#define ERRBUFSIZE 1024
void            fatal(msg)
    char           *msg;
{
    char           *prefix = "ASAP : ";

    write(2, prefix, strlen(prefix));
    write(2, msg, strlen(msg));
    write(2, "\n", 1);
    exit(1);
}

/*
 * "GetHostAddress (buf)" returns the internet address of the host we are
 * running on as a string in the x.x.x.x form in "buf". "buf" should be thus
 * at least 16 characters long Inability to find this address is fatal.
 */

void            GetLocalHostAddress(buf)
    char           *buf;
{
    char            hostname[MAXHOSTNAMELEN + 1];
    struct hostent *hostent;

    buf[0] = '\0';
    if (gethostname(hostname, MAXHOSTNAMELEN + 1) < 0) {
	perror("cannot find local host name");
	exit(2);
    }

    if ((hostent = gethostbyname(hostname)) == (struct hostent *) 0) {
	fatal("cannot find local host address");
    }


    {
	unsigned long  *paddr, horder;

	paddr = *hostent->h_addr_list;
	horder = ntohl(*paddr);
	sprintf(buf, "%u.%u.%u.%u", horder >> 24, (horder >> 16) & 0xff,
		(horder >> 8) & 0xff, horder & 0xff);
    }

}


void            initConn(conn, nsd, usd, path)
    ASAPconn_t     *conn;
    int             nsd, usd;
    char           *path;
{
    channel_t       normal;
    channel_t       urgent;

    normal = (channel_t) ASAPalloc(sizeof(struct channel_struct));
    urgent = (channel_t) ASAPalloc(sizeof(struct channel_struct));

    /* necessary to force the first read ... */
    normal->in.last = -1;
    urgent->in.last = -1;

    normal->in.fd = nsd;
    normal->out.fd = nsd;
    normal->in.buf = ASAPalloc((unsigned) ASAPBUFSIZ);
    normal->out.buf = ASAPalloc((unsigned) ASAPBUFSIZ);
    normal->in.size = ASAPBUFSIZ;
    normal->out.size = ASAPBUFSIZ;
    normal->in.curr = 0;

    normal->out.curr = 0;
    normal->in.tok = ASAP_UNKNOWN;

    urgent->in.fd = usd;
    urgent->out.fd = usd;
    urgent->in.buf = ASAPalloc((unsigned) ASAPUBUFSIZ);
    urgent->out.buf = ASAPalloc((unsigned) ASAPUBUFSIZ);
    urgent->in.size = ASAPUBUFSIZ;
    urgent->out.size = ASAPUBUFSIZ;
    urgent->in.curr = 0;
    urgent->out.curr = 0;
    urgent->in.tok = ASAP_UNKNOWN;

    conn->normal = normal;
    conn->urgent = urgent;
    /* Do we need to store the pathname of the UNIX socket ? */
    if (path != (char *) 0) {
	conn->path = (char *) ASAPalloc(strlen(path) + 1);
	strcpy(conn->path, path);
    }
    else
	conn->path = (char *) 0;


}

int             encodeInt(buf, i)
    char           *buf;
    int             i;
{
    int             p;
    int             cur;

    char           *digits = "0123456789abcdef";

    p = (i < 0) ? -i : i;

    if (p < 10) {
	buf[0] = digits[p];
	cur = 0;
    }
    else {
	buf[0] = '0';
	buf[1] = 'x';
	cur = 1;
	do {
	    buf[++cur] = digits[p % 16];
	} while (p = p / 16);
    }

    if (i < 0) {
	buf[++cur] = '-';
	return (cur + 1);
    }
    else {
	return (cur + 1);
    }

}
