#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <varargs.h>
#include <sys/errno.h>		/* for error code during open */

#include "myAll.h"
#include "myError.h"

/* any call to myExit is fatal... segmentation fault (core dumped)!! */

void            myExit(int status)
{
#ifdef NO_EXIT
    fflush(stdout);
    MAKECORE			/* What a nice core dump!!! */
#endif
	exit(status);
}

void            warn(char *wm)
{
    fprintf(stderr, "Warning: %s\n", wm);
}

void            warn2(char *wm1, char *wm2)
{
    fprintf(stderr, "Warning: %s%s\n", wm1, wm2);
}

void            warn3(char *wm1, char *wm2, char *wm3)
{
    fprintf(stderr, "Warning: %s%s%s\n", wm1, wm2, wm3);
}

void            warn4(char *wm1, char *wm2, char *wm3, char *wm4)
{
    fprintf(stderr, "Warning: %s%s%s%s\n", wm1, wm2, wm3, wm4);
}



void            error(char *erm)
{
    fprintf(stderr, "Error: %s\n", erm);
    myExit(1);
}

void            error2(char *erm1, char *erm2)
{
    fprintf(stderr, "Error: %s%s\n", erm1, erm2);
    myExit(1);
}

void            error3(char *erm1, char *erm2, char *erm3)
{
    fprintf(stderr, "Error: %s%s%s\n", erm1, erm2, erm3);
    myExit(1);
}

void            error4(char *erm1, char *erm2, char *erm3, char *erm4)
{
    fprintf(stderr, "Error: %s%s%s%s\n", erm1, erm2, erm3, erm4);
    myExit(1);
}



char           *openError(int err, char *f)
{
    char            tmp[MAX_CHAR_TMP];
    char            tmp2[MAX_CHAR_TMP];

    switch (err) {
    case EACCES:{
	    sprintf(tmp2, "file not found or permision denied (EACCES)");
	    break;
	}
    case EDQUOT:{
	    sprintf(tmp2, "quota overflow (EDQUOT)");
	    break;
	}
    case EEXIST:{
	    sprintf(tmp2, "file already exist (EEXIST)");
	    break;
	}
    case EFAULT:{
	    sprintf(tmp2, "mem error (EFAULT)");
	    break;
	}
    case EINTR:{
	    sprintf(tmp2, "signal caught during open (EINTR)");
	    break;
	}
    case EIO:{
	    sprintf(tmp2, "hangup error (EIO)");
	    break;
	}
    case EISDIR:{
	    sprintf(tmp2, "it's a directory (EISDIR)");
	    break;
	}
    case ELOOP:{
	    sprintf(tmp2, "too many symbolic links (ELOOP)");
	    break;
	}
    case EMFILE:{
	    sprintf(tmp2, "too much files have already been opened by this process (EMFILE)");
	    break;
	}
    case ENAMETOOLONG:{
	    sprintf(tmp2, "too long path name (ENAMETOOLONG)");
	    break;
	}
    case ENFILE:{
	    sprintf(tmp2, "too much files have already been opened (ENFILE)");
	    break;
	}
    case ENOENT:{
	    sprintf(tmp2, "file does not exist (ENOENT)");
	    break;
	}
    case ENOSPC:{
	    sprintf(tmp2, "disk full (ENOSPC)");
	    break;
	}
    case ENOTDIR:{
	    sprintf(tmp2, "component of the path prefix is not a directory (ENOTDIR)");
	    break;
	}
    case ENXIO:{
	    sprintf(tmp2, "(ENXIO)");
	    break;
	}
    case EROFS:{
	    sprintf(tmp2, "read only file system (EROFS)");
	    break;
	}
    default:{
	    sprintf(tmp2, "unknown error (%d)", err);
	    break;
	}
    }
    sprintf(tmp, "can't open file %s: %s", f, tmp2);
    return myStrDup(tmp);
}

/*
 * A wrapper for exit
 * if NO_EXIT is defined any call to UExit is fatal...
 *  segmentation fault (core dumped)!!
 * useful when debuginig
 */
void
UExit(int status)
{
#ifdef NO_EXIT
    fflush(stdout);
    MAKECORE                            /* What a nice core dump!!! */
#endif
        exit(status);
}


/*
 * UIWarning should be called like:
 *    UIWarning(function_name, format, arg1, arg2...);
 *    the format string is like this used by printf
 *    ex:
 *         UIWarning("CheckedMalloc", "can't allocate %d bytes, %d allocated"
 *                       ,size, allocatedSize);
 */
void
UIWarning(char * f)
{
    /* print name of function causing warning */
    (void) fprintf(stderr, "Internal WARNING in %s: ", f);
}


/*
 * UIError should be called like:
 *    UIError(function_name, format, arg1, arg2...);
 *    the format string is like this used by printf
 *    ex:
 *         UIError("CheckedMalloc", "can't allocate %d bytes, %d allocated"
 *                       ,size, allocatedSize);
 */
void
UIError(char * f)
{
    (void) fprintf(stderr, "Internal ERROR in %s: ", f);
    UExit(1);
}
