#include <stdio.h>
#include <string.h>
#include <memory.h>
#include <malloc.h>

#include "myAll.h"
#include "limits.h"
#include "myStruct.h"

static OGNode  *ogCreateNode(OGraph * pog, void *pe);
static void     ogFreeNode(OGNode * pgn);
static void     ogAddSucc(OGNode * pgna, OGNode * pgnb);
static void    *ogKeyFunc(OGNode * pgn);
static void     ogPrintFunc(OGNode * pgn);
static void     ogSetFlags(OGraph * pog, int flag);
static void     ogFindTreeRootAux(OGNode * pogn, Tree * pt, TreeNode * ptnf);
static Tree    *ogFindTreeRoot(OGNode * pogn);
static void     tFreeSons(TreeNode * ptn, unsigned char freeElt);
static void     tAddSon(TreeNode * ptnf, TreeNode * ptn);
static List    *tbfb(TreeNode * ptn, List * pl);
static List    *tdfb(TreeNode * ptn, List * pl);
static void     tPrintAux(TreeNode * ptn);
static void     lfreeListElt(List * pl, ListElt * ple, unsigned char freeElt);




/***************************************************************/
/* some predefined functions usefull for hash, list, tree, ... */

int             hashString(char *str, int sz)
{
    if (str == (char *) NULL)
	error("hashStringFunc: NULL string ");
    return str[0] % sz;
}

int             hashInt(int *i, int sz)
{
    return abs((*i * 9973) % sz);
}

int             hashVoid(void *p, int sz)
{
    return abs(((unsigned long) p * 9973) % sz);
}


int             ordString(char *s1, char *s2)
{
    return strcmp(s1, s2);
}

int             ordInt(int *a, int *b)
{
    return (*a > *b) ? 1 : (*a == *b) ? 0 : -1;
}

int             ordVoid(void *a, void *b)
{
    return (a > b) ? 1 : (a == b) ? 0 : -1;
}

void            printString(char *pe)
{
    puts(pe);
}

void            printInt(int *pe)
{
    printf("%d", *pe);
}

void            printVoid(void *pe)
{
    printf("(void*)%x", pe);
}

void           *keySame(void *pe)
{
    return pe;
}

void           *keyVoid(void *pe)
{
    return pe;
}


HashTable      *htCreate(int size, void *(*keyFunc) (), int (*orderFunc) (), void (*freeFunc) (), int (*hashFunc) (), void (*printFunc) ())
{
    HashTable      *ht;
    static          number = 1;

    if ((keyFunc == NULL) || (orderFunc == NULL) || (freeFunc == NULL) || (hashFunc == NULL) || (size <= 0))
	error("createHashTable: bad parameters");
    ht = (HashTable *) myMallocMsg(sizeof(HashTable), "createHashTable: ");
    ht->size = size;
    ht->nbe = 0;
    ht->number = number++;
    ht->modified = True;
    ht->lastAskedn = -1;
    ht->lastHTInd = 0;
    ht->lastEltCount = 0;
    ht->hla = (HashListHeader *) myCallocMsg(size, sizeof(HashListHeader), "createHashTable: ");
    ht->keyFunc = keyFunc;
    ht->ordFunc = orderFunc;
    ht->freeFunc = freeFunc;
    ht->hashFunc = hashFunc;
    if (printFunc == NULL)
	printFunc = printVoid;
    ht->printFunc = printFunc;
    return ht;
}

void            htFree(HashTable * pht, unsigned char freeElt)
{
    int             i;
    HashList       *hl, *hlb;

    for (i = 0; i < pht->size; i++) {
	hl = pht->hla[i].l;
	while (hl != NULL) {
	    if (freeElt)
		(*pht->freeFunc) (hl->pe);
	    hlb = hl;
	    hl = hl->next;
	    free(hlb);
	}
    }
    free(pht->hla);
    free(pht);
}

void           *htSearch(HashTable * pht, void *pe)
{
    int             hind;
    HashList       *hl;
    void           *k1, *k2;

    k2 = (*pht->keyFunc) (pe);
    hind = (*pht->hashFunc) (k2, pht->size);
    hl = pht->hla[hind].l;
    while (hl != NULL) {
	k1 = (*pht->keyFunc) (hl->pe);
	if (!(*pht->ordFunc) (k1, k2))
	    return hl->pe;
	hl = hl->next;
    }
    return NULL;
}

void           *htSearchKey(HashTable * pht, void *key)
{
    int             hind;
    HashList       *hl;
    void           *k1;

    hind = (*pht->hashFunc) (key, pht->size);
    hl = pht->hla[hind].l;
    while (hl != NULL) {
	k1 = (*(pht->keyFunc)) (hl->pe);
	if (((*pht->ordFunc) (k1, key)) == 0)
	    return hl->pe;
	hl = hl->next;
    }
    return NULL;
}

void            htInsert(HashTable * pht, void *pe)
{
    int             hind;
    HashList       *hl;

    pht->modified = True;
    hind = (*pht->hashFunc) ((*pht->keyFunc) (pe), pht->size);
    if (htSearch(pht, pe) == NULL) {
	hl = (HashList *) myMallocMsg(sizeof(HashList), "htInsert: ");
	hl->pe = pe;
	hl->next = pht->hla[hind].l;
	pht->hla[hind].l = hl;
	pht->hla[hind].lsz++;
	pht->nbe++;
    }
}


void            htDel(HashTable * pht, void *pe, unsigned char freeElt)
{
    int             hind;
    HashList       *hl, *hlp;
    void           *k1, *k2;

    pht->modified = True;
    k2 = (*pht->keyFunc) (pe);
    hind = (*pht->hashFunc) (k2, pht->size);
    hl = pht->hla[hind].l;
    while (hl != NULL) {
	k1 = (*pht->keyFunc) (hl->pe);
	if (!(*pht->ordFunc) (k1, k2)) {
	    if (hl == pht->hla[hind].l)
		pht->hla[hind].l = hl->next;
	    else
		hlp->next = hl->next;
	    if (freeElt)
		free(hl->pe);
	    free(hl);
	    pht->hla[hind].lsz--;
            return;
	}
	hlp = hl;
	hl = hl->next;
    }
}


int             htNbElts(HashTable * pht)
{
    return pht->nbe;
}

void           *htFindNth(HashTable * pht, int n)
{
    HashList       *hl;
    int             nbelt;
    int             i;
    int             htInd;


    if (n <= 0)
	error("htFindNth: bad element rank (<=0)");
    if ((n >= pht->lastAskedn) && (pht->modified == False)) {
	nbelt = pht->lastEltCount;
	htInd = pht->lastHTInd;
    }
    else {
	nbelt = 0;
	htInd = 0;
    }
    pht->modified = False;
    pht->lastAskedn = n;
    for (i = htInd; i < pht->size; i++) {
	if (nbelt + pht->hla[i].lsz >= n) {
	    hl = pht->hla[i].l;
	    pht->lastEltCount = nbelt;
	    while (nbelt < n - 1) {
		nbelt++;
		hl = hl->next;
	    }
	    pht->lastHTInd = i;
	    return hl->pe;
	}
	nbelt += pht->hla[i].lsz;
    }
    error("htFindNth: no element of this rank ");
    return NULL;		/* not reached */
}




/******************/
/* Oriented Graph */

static OGNode  *ogCreateNode(OGraph * pog, void *pe)
{
    OGNode         *pgn;

    pgn = (OGNode *) myMallocMsg(sizeof(OGNode), "new node for Oriented Graph");
    pgn->pe = pe;
    pgn->flag = 0;
    pgn->nbs = 0;
    pgn->nbp = 0;
    pgn->sa = NULL;
    pgn->pog = pog;
    return pgn;
}

static void     ogFreeNode(OGNode * pgn)
{
    (*(pgn->pog->freeFunc)) (pgn->pe);
    free(pgn);
}

static void     ogAddSucc(OGNode * pgna, OGNode * pgnb)
{
    if ((pgna->nbs % OGNSUCC_CHKSZ) == 0) {
	if (pgna->sa == NULL)
	    pgna->sa = (OGNode **) myMalloc(OGNSUCC_CHKSZ * sizeof(OGNode *));
	else
	    pgna->sa = (OGNode **) myRealloc(pgna->sa, (pgna->nbs + OGNSUCC_CHKSZ) * sizeof(OGNode *));
    }
    pgna->sa[pgna->nbs] = pgnb;
    pgnb->nbp++;
    pgna->nbs++;
}

static void    *ogKeyFunc(OGNode * pgn)
{
    return (*(pgn->pog->keyFunc)) (pgn->pe);
}

static void     ogPrintFunc(OGNode * pgn)
{
    (*(pgn->pog->printFunc)) (pgn->pe);
}


OGraph         *ogCreate(int orderOfSize, void *(*keyFunc) (), int (*orderFunc) (), void (*freeFunc) (), int (*hashFunc) (), void (*printFunc) ())
{
    OGraph         *pg;

    if ((keyFunc == NULL) || (orderFunc == NULL) || (freeFunc == NULL) || (hashFunc == NULL))
	error("createOGraph: bad parameters");
    pg = (OGraph *) myMallocMsg(sizeof(OGraph), "createOGraph: ");
    pg->es = orderOfSize;
    pg->nbn = 0;
    pg->keyFunc = keyFunc;
    pg->ordFunc = orderFunc;
    pg->freeFunc = freeFunc;
    pg->hashFunc = hashFunc;
    if (printFunc == NULL)
	printFunc = printVoid;
    pg->printFunc = printFunc;
    pg->ht = htCreate(pg->es / 10 + 1, ogKeyFunc, orderFunc, ogFreeNode, hashFunc, ogPrintFunc);
    return pg;
}


void            ogFree(OGraph * pog)
{
    OGNode         *pogn;
    int             i;

    for (i = 1; i <= pog->nbn; i++) {
	pogn = htFindNth(pog->ht, i);
	free(pogn->sa);
    }
    htFree(pog->ht, True);
    free(pog);
}


void            ogAddNode(OGraph * pog, void *pe)
{
    OGNode          tmp;

    tmp.pog = pog;		/* simulate a nod for compatibility with
				 * ogKeyFunc */
    tmp.pe = pe;
    if (!(htSearch(pog->ht, &tmp))) {
	htInsert(pog->ht, (void *) ogCreateNode(pog, pe));
	pog->nbn++;
    }
}


/* add an arc between node pa and pb (pa->pb) */
/* nodes are created if needed */
void            ogAddArc(OGraph * pog, void *pa, void *pb)
{
    OGNode         *pgna, *pgnb;
    OGNode          tmp;

    tmp.pog = pog;
    tmp.pe = pa;
    if (!(pgna = htSearch(pog->ht, &tmp))) {
	pgna = ogCreateNode(pog, pa);
	htInsert(pog->ht, (void *) pgna);
	pog->nbn++;
    }
    tmp.pe = pb;
    if (!(pgnb = htSearch(pog->ht, &tmp))) {
	pgnb = ogCreateNode(pog, pb);
	htInsert(pog->ht, (void *) pgnb);
	pog->nbn++;
    }
    ogAddSucc(pgna, pgnb);
}

static void     ogSetFlags(OGraph * pog, int flag)
{
    int             i;
    OGNode         *pogn;

    for (i = 1; i <= pog->nbn; i++) {
	pogn = (OGNode *) htFindNth(pog->ht, i);
	pogn->flag = flag;
    }
}


static void     ogFindTreeRootAux(OGNode * pogn, Tree * pt, TreeNode * ptnf)
{
    TreeNode       *ptn;
    int             i;

    if (pogn->flag == 0) {
	ptn = tAddNode(pt, ptnf, pogn->pe);
	pogn->flag = 1;
	for (i = 0; i < pogn->nbs; i++) {
	    ogFindTreeRootAux(pogn->sa[i], pt, ptn);
	}
    }
}

static Tree    *ogFindTreeRoot(OGNode * pogn)
{
    Tree           *pt;

    pt = tCreate(pogn->pog->keyFunc, pogn->pog->ordFunc, pogn->pog->freeFunc, pogn->pog->printFunc);
    ogSetFlags(pogn->pog, 0);
    ogFindTreeRootAux(pogn, pt, NULL);
    return pt;
}

Boolean         ogCheckCycleAux(OGNode * pogn)
{
    int             i;

    if (pogn->flag == 2)
	return True;		/* this node is marked, so it's is own
				 * ancester. cycle! */
    if (pogn->flag == 1)
	return False;		/* this is just an old one. it's already
				 * checked */
    pogn->flag = 2;		/* now it's an ancester */
    for (i = 0; i < pogn->nbs; i++) {	/* let's check these sons */
	if (ogCheckCycleAux(pogn->sa[i]) == True)
	    return True;
    }
    pogn->flag = 1;		/* all this branch has been checked */
    return False;
}

Boolean         ogCheckCycle(OGraph * pog)
{
    int             i;
    OGNode         *pogn;

    for (i = 1; i <= pog->nbn; i++) {
	pogn = htFindNth(pog->ht, i);
	if (pogn->flag == 1)
	    continue;		/* this node has already been checked */
	if (ogCheckCycleAux(pogn) == True)
	    return True;
    }
    return False;
}


/* find an order in an oriented graph                                      */
/* the returned list is ordered from most prior to less prior in the graph */
/* for instance:                                                           */
/* graph:   a-> b ->c     list:  a->b->c->d  or  a->d->b->c                */
/* \> d                                                          */
/* Warning: graph must not be cyclic                                       */
List           *ogFindOrder(OGraph * pog)
{
    int             i, j, k;
    List           *pl;
    OGNode        **aogn;

    pl = lCreate(pog->keyFunc, pog->ordFunc, pog->freeFunc, pog->printFunc);
    aogn = (OGNode **) myMalloc(pog->nbn * sizeof(OGNode *));
    for (i = 0; i < pog->nbn; i++) {
	aogn[i] = htFindNth(pog->ht, i + 1);	/* build a fast access struct
						 * for nodes */
	aogn[i]->flag = aogn[i]->nbp;
    }
    for (i = 0; i < pog->nbn; i++) {	/* to order all the nodes */
	for (j = 0; j < pog->nbn - i; j++) {	/* the unordered nodes are in
						 * the head of aogn */
	    if (aogn[j]->flag == 0) {	/* this node has no predecessor */
		lAdd(pl, aogn[j]->pe);	/* take this node. it's the next in
					 * this order */
		for (k = 0; k < aogn[j]->nbs; k++)	/* cut it off the graph */
		    aogn[j]->sa[k]->flag--;	/* it's no longer a
						 * predecessor */
		aogn[j] = aogn[(pog->nbn) - i - 1];	/* to keep the unordered
							 * nodes in the head */
		break;
	    }
	}
    }
    free(aogn);
    lReverse(pl);
    return pl;
}

Tree           *ogFindTree(OGraph * pog)
{
    int             i;
    Tree           *pt;

    for (i = 1; i <= pog->nbn; i++) {
	pt = ogFindTreeRoot(htFindNth(pog->ht, i));
	if (pt->nbn == pog->nbn)
	    return pt;
	else
	    tFree(pt, False);
    }
    return NULL;
}





/********/
/* Tree */

Tree           *tCreate(void *(*keyFunc) (), int (*orderFunc) (), void (*freeFunc) (), void (*printFunc) ())
{
    Tree           *pt;

    if ((keyFunc == NULL) || (orderFunc == NULL) || (freeFunc == NULL))
	error("createTree: bad parameters");
    pt = (Tree *) myMalloc(sizeof(Tree));
    pt->nbn = 0;
    pt->keyFunc = keyFunc;
    pt->ordFunc = orderFunc;
    pt->freeFunc = freeFunc;
    if (printFunc == NULL)
	printFunc = printVoid;
    pt->printFunc = printFunc;
    pt->root = NULL;
    return pt;
}

static void     tFreeSons(TreeNode * ptn, unsigned char freeElt)
{
    int             i;

    if (ptn == NULL)
	return;
    for (i = 0; i < ptn->nbs; i++)
	tFreeSons(ptn->sa[i], freeElt);
    if (freeElt)
	(*(ptn->pt->freeFunc)) (ptn->pe);
    free(ptn->sa);
    free(ptn);
}

void            tFree(Tree * pt, unsigned char freeElt)
{
    tFreeSons(pt->root, freeElt);
    free(pt);
}

static void     tAddSon(TreeNode * ptnf, TreeNode * ptn)
{
    if ((ptnf->nbs % TREESA_CHKSZ) == 0) {
	if (ptnf->sa == NULL)
	    ptnf->sa = (TreeNode **) myMalloc(TREESA_CHKSZ * sizeof(TreeNode *));
	else
	    ptnf->sa = (TreeNode **) myRealloc(ptnf->sa, (ptnf->nbs + TREESA_CHKSZ) * sizeof(TreeNode *));
    }
    ptnf->sa[ptnf->nbs] = ptn;
    ptnf->nbs++;
}

TreeNode       *tAddNode(Tree * pt, TreeNode * ptnf, void *pe)
{
    TreeNode       *ptn;

    if ((ptnf != NULL) && (pt != ptnf->pt))
	error("tAddNode mismatched tree and treeNode");
    ptn = (TreeNode *) myMallocMsg(sizeof(TreeNode), "alloc a new tree node");
    ptn->pe = pe;
    ptn->flag = 0;
    ptn->nbs = 0;
    ptn->sa = NULL;
    ptn->pt = pt;
    if ((ptnf == NULL) && (pt->root != NULL))
	error("tAddNode: alerady a root for this tree!");
    if (pt->root == NULL)
	pt->root = ptn;
    else
	tAddSon(ptnf, ptn);
    pt->nbn++;
    return ptn;
}

int             tNbSons(TreeNode * ptn)
{
    return ptn->nbs;
}

static List    *tbfb(TreeNode * ptn, List * pl)
{
    int             i;

    for (i = 0; i < ptn->nbs; i++) {
	lAdd(pl, ptn->sa[i]->pe);
    }
    for (i = 0; i < ptn->nbs; i++) {
	tbfb(ptn->sa[i], pl);
    }
    return pl;
}

List           *tBreadthFirst(Tree * pt)
{
    List           *pl;

    pl = lCreate(pt->keyFunc, pt->ordFunc, pt->freeFunc, pt->printFunc);
    if (pt->root == NULL)
	return pl;
    lAdd(pl, pt->root->pe);
    pl = tbfb(pt->root, pl);
    lReverse(pl);
    return pl;
}

static List    *tdfb(TreeNode * ptn, List * pl)
{
    int             i;

    for (i = 0; i < ptn->nbs; i++) {
	tdfb(ptn->sa[i], pl);
    }
    lAdd(pl, ptn->pe);
    return pl;
}

List           *tDephtFirst(Tree * pt)
{
    List           *pl;

    pl = lCreate(pt->keyFunc, pt->ordFunc, pt->freeFunc, pt->printFunc);
    if (pt->root == NULL)
	return pl;
    pl = tdfb(pt->root, pl);
    lReverse(pl);
    return pl;
}

static void     tPrintAux(TreeNode * ptn)
{
    static int      indent = 0;
    int             i;

    for (i = 0; i < indent; i++)
	putchar(' ');
    indent += 3;
    (*(ptn->pt->printFunc)) (ptn->pe);
    putchar('\n');
    for (i = 0; i < ptn->nbs; i++)
	tPrintAux(ptn->sa[i]);
    indent -= 3;
}

void            tPrint(Tree * pt)
{
    tPrintAux(pt->root);
}


/********/
/* List */

List           *lCreate(void *(*keyFunc) (), int (*orderFunc) (), void (*freeFunc) (), void (*printFunc) ())
{
    List           *pl;
    static          number = 1;

    if ((keyFunc == NULL) || (orderFunc == NULL) || (freeFunc == NULL))
	error("createList: bad parameters");
    pl = (List *) myMalloc(sizeof(List));
    pl->nbe = 0;
    pl->number = number++;
    pl->modified = True;
    pl->lastN = 0;
    pl->lastAns = NULL;
    pl->keyFunc = keyFunc;
    pl->ordFunc = orderFunc;
    pl->freeFunc = freeFunc;
    if (printFunc == NULL)
	printFunc = printVoid;
    pl->printFunc = printFunc;
    pl->first = NULL;
    return pl;
}

static void     lfreeListElt(List * pl, ListElt * ple, unsigned char freeElt)
{
    if (freeElt) {
	(*(pl->freeFunc)) (ple->pe);
    }
    free(ple);
}

void            lFree(List * pl, unsigned char freeElt)
{
    ListElt        *ple, *pleb;

    ple = pl->first;
    while (ple) {
	pleb = ple;
	ple = ple->next;
	lfreeListElt(pl, pleb, freeElt);
    }
    free(pl);
}

void            lAdd(List * pl, void *pe)
{
    ListElt        *ple;

    pl->modified = True;
    ple = (ListElt *) myMallocMsg(sizeof(ListElt), "alloc a new list element ");
    ple->pe = pe;
    ple->flag = 0;
    ple->next = pl->first;
    pl->first = ple;
    pl->nbe++;
}

void            lAddUnique(List * pl, void *pe)
{
    ListElt        *ple;

    /* to be optimized */

    if (lIn(pl, pe) == False)
	lAdd(pl, pe);
}

int             lNbElts(List * pl)
{
    return pl->nbe;
}

void           *lFindNth(List * pl, int n)
{
    int             nb;
    ListElt        *ple;

    if ((n >= pl->lastN) && (pl->modified == False) && (pl->lastAns != NULL)) {
	nb = pl->lastN;
	ple = pl->lastAns;
    }
    else {
	nb = 1;
	ple = pl->first;
    }
    pl->modified = False;
    while (ple != NULL) {
	if (nb == n) {
	    pl->lastAns = ple;
	    pl->lastN = n;
	    return ple->pe;
	}
	ple = ple->next;
	nb++;
    }
    error("lFindNth: no element of this rank ");
    return NULL;		/* not reached */
}


void            lDel(List * pl, void *pe, unsigned char freeElt)
{
    ListElt        *ple, *plb = NULL;

    pl->modified = True;
    ple = pl->first;
    while (ple != NULL) {
	if (pl->ordFunc(ple->pe, pe) == 0) {
	    if (plb == NULL)
		pl->first = ple->next;
	    else
		plb->next = ple->next;
	    pl->nbe--;
	    lfreeListElt(pl, ple, freeElt);
	    break;
	}
	plb = ple;
	ple = ple->next;
    }
}


void            lDelAll(List * pl, void *pe, unsigned char freeElt)
{
    ListElt        *ple, *plb = NULL;

    pl->modified = True;
    ple = pl->first;
    while (ple != NULL) {
	if (pl->ordFunc(ple->pe, pe) == 0) {
	    if (plb == NULL)
		pl->first = ple->next;
	    else
		plb->next = ple->next;
	    pl->nbe;
	    ple = ple->next;
	    lfreeListElt(pl, ple, freeElt);
	}
	else {
	    plb = ple;
	    ple = ple->next;
	}
    }
}


/*
 * return the first ListElt* in which pe match key pk
 */
static
ListElt *
lLookListKey(List * pl, void *pk)
{
    ListElt *ple;

    ple = pl->first;
    while (ple != NULL) {
        if (pl->ordFunc(pl->keyFunc(ple->pe), pk) == 0) {
            return ple;
        }
        ple = ple->next;
    }
    return NULL;
}


Boolean         lIn(List * pl, void *pe)
{
    ListElt        *ple;

    ple = pl->first;
    while (ple != NULL) {
	if (pl->ordFunc(ple->pe, pe) == 0) {
	    return True;
	}
	ple = ple->next;
    }
    return False;
}


/*
 * return first element in pl matching key pk
 */
void *
lLookKey(List * pl, void *pk)
{
    ListElt *ple;

    return (ple = lLookListKey(pl, pk)) ? ple->pe : NULL;
}


void           *lSearch(List * pl, void *pe)
{
    ListElt        *ple;

    ple = pl->first;
    while (ple != NULL) {
	if (pl->ordFunc(ple->pe, pe) == 0) {
	    return ple->pe;
	}
	ple = ple->next;
    }
    return NULL;
}

void            lReverse(List * pl)
{
    ListElt        *p, *pp, *pn;

    pl->modified = True;
    p = pl->first;
    pp = NULL;
    while (p != NULL) {
	pn = p->next;
	p->next = pp;
	pp = p;
	p = pn;
    }
    pl->first = pp;
}


void            lPrint(List * pl)
{
    ListElt        *p;

    printf("((%d :%d elts at x%x):\n  ", pl->number, pl->nbe, pl);
    p = pl->first;
    while (p != NULL) {
	(*(pl->printFunc)) (p->pe);
	p = p->next;
	if (p != NULL)
	    printf("->");
    }
    printf(")\n");
}




/*********/
/* Stack */

Stack          *sCreate(void)
{
    Stack          *ps;

    ps = (Stack *) myMalloc(sizeof(Stack));
    ps->nbe = 0;
    ps->sz = 0;
    ps->ea = NULL;
    return ps;
}

void            sFree(Stack * ps)
{
    free(ps->ea);
    free(ps);
}

Boolean         sIsEmpty(Stack * ps)
{
    return (ps->nbe == 0) ? True : False;
}

void           *sPop(Stack * ps)
{
    if (ps->nbe > 0) {
	ps->nbe--;
	return ps->ea[ps->nbe];
    }
    else
	return NULL;
}

void            sPush(Stack * ps, void *pe)
{
    ps->nbe++;
    if (ps->nbe > ps->sz) {
	if (ps->ea == NULL) {
	    ps->sz = STACK_CHKSZ;
	    ps->ea = (void **) myMallocMsg(ps->sz * sizeof(void *), "Stack allocation ");
	}
	else {
	    ps->sz += STACK_CHKSZ;
	    ps->ea = (void **) myReallocMsg(ps->ea, ps->sz * sizeof(void *), "Stack enlargment ");
	}
    }
    ps->ea[ps->nbe - 1] = pe;
}

void           *sLook(Stack * ps)
{
    return ps->ea[ps->nbe - 1];
}




/*********/
/* Queue */

Queue          *qCreate(void)
{
    Queue          *pq;

    pq = (Queue *) myMalloc(sizeof(Queue));
    pq->nbe = 0;
    pq->sz = QUEUE_STARTSZ;
    pq->ea = (void **) myMalloc(pq->sz * sizeof(void *));
    pq->iend = 0;
    pq->ihead = 0;
    pq->firstNull = INT_MAX;
    return pq;
}


void            qFree(Queue * pq)
{
    free(pq->ea);
    free(pq);
}

static int      QPos(Queue * pq, int pos)
{
    int             np;

    np = pos + pq->ihead;
    if (np > pq->sz)
	return np - pq->sz;
    else if (np < 0)
	return np + pq->sz;
    else
	return np;
}


void            qAdd(Queue * pq, void *pe)
{
    void          **eatmp;

    if (pq->nbe == pq->sz) {
	pq->sz = (float) pq->sz * QUEUE_MULTSZ + 1;
	if (pq->iend > pq->ihead)
	    pq->ea = (void **) myReallocMsg(pq->ea, pq->sz * sizeof(void *), "Queue can't grow");
	else {
	    eatmp = (void **) myMallocMsg(pq->sz * sizeof(void *), "Queue can't grow");
	    memcpy(eatmp, &pq->ea[pq->ihead], (pq->nbe - pq->ihead) * sizeof(void *));
	    memcpy(&eatmp[pq->nbe - pq->ihead], pq->ea, pq->ihead * sizeof(void *));
	    free(pq->ea);
	    pq->ea = eatmp;
	    pq->ihead = 0;
	    pq->iend = pq->nbe;
	}
    }
    pq->ea[pq->iend] = pe;
    pq->iend = (pq->iend + 1) % pq->sz;
    pq->nbe++;
}

void           *qGet(Queue * pq)
{
    void           *pe;

    if (pq->nbe <= 0)
	return NULL;
    pe = pq->ea[pq->ihead];
    pq->ihead = (pq->ihead + 1) % pq->sz;
    pq->nbe--;
    return pe;
}

void            qDel(Queue * pq, int nb, void (*freeFunc) ())
{
    int             i;

    for (i = 0; i < nb; i++) {
	if (pq->nbe <= 0)
	    return;
	(*freeFunc) (pq->ea[pq->ihead]);
	pq->ihead = (pq->ihead + 1) % pq->sz;
	pq->nbe--;
    }
}

void           *qLook(Queue * pq)
{
    return pq->ea[pq->ihead];
}

void           *qLookNb(Queue * pq, int nb)
{
    int             ind;

    ind = (pq->ihead + nb - 1) % pq->sz;
    return pq->ea[ind];
}


void           *qShift(Queue * pq, int start, int shift)
{
    int             i, id, is;

    if (shift > 0) {
	fprintf(stderr, "error: qShift: for positive numbers (%d) not implemented\n", shift);
	exit(1);
    }
    if ((start + shift) < 0) {
	fprintf(stderr, "warning: qShift: loosing some values (%d) due to a too big shift\n", shift + start);
	start = -shift;
    }
    if (start > pq->nbe) {
	fprintf(stderr, "warning: qShift: shift has no effect, start to high (%d)\n", start);
    }
    /* -- convert the user index [1->nbelt] to the real index [0->nbelt-1] */
    start--;
    for (i = start; i < pq->nbe; i++) {
	id = QPos(pq, i + shift);
	is = QPos(pq, i);
	pq->ea[id] = pq->ea[is];
    }
    pq->nbe += shift;
    pq->iend = QPos(pq, pq->nbe);
}


int             qNbElts(Queue * pq)
{
    return pq->nbe;
}

void            qNullNb(Queue * pq, int nb)
{
    pq->ea[QPos(pq, nb - 1)] = NULL;
    if (pq->firstNull > nb)
	pq->firstNull = nb - 1;
}

void            qCompactNull(Queue * pq)
{
    int             ie, ic, nbe, i;

    if (pq->firstNull > pq->nbe)
	return;
    ie = ic = QPos(pq, pq->firstNull);
    nbe = pq->nbe;
    for (i = pq->firstNull; i < pq->nbe; i++) {
	if (pq->ea[ic]) {
	    if (ie != ic)	/* avoid useless copy */
		pq->ea[ie] = pq->ea[ic];
	    ie++;
	    if (ie >= pq->sz)
		ie = 0;
	}
	else {
	    nbe--;
	}
	ic++;
	if (ic >= pq->sz)
	    ic = 0;
    }
    pq->iend = ie;
    pq->nbe = nbe;
    pq->firstNull = INT_MAX;
}





/********************************************/
/* a very frequently used list: string list */

LString        *lsAdd(LString * ls, char *s)
{
    LString        *ps;

    if (s == NULL)
	return ls;
    ps = (LString *) myMalloc(sizeof(LString));
    ps->next = ls;
    ps->s = s;
    return ps;
}

Boolean         lsIn(LString * ls, char *s)
{
    while (ls) {
	if (!strcmp(ls->s, s))
	    return True;
	ls = ls->next;
    }
    return False;
}

void            lsFree(LString * ls, unsigned char freeElt)
{
    LString        *ps;

    while (ls != NULL) {
	if (freeElt)
	    free(ls->s);
	ps = ls->next;
	free(ls);
	ls = ps;
    }
}
