#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/time.h>

#include "asap.h"
#include "agattypes.h"
#include "myMisc.h"

#include "agatcheck.h"
#include "agatparse.h"
#include "agatbuild.h"
#include "agatprocess.h"
#include "agatXgen.h"
#include "agat.h"


typedef struct DescFid {
    char           *name;	/* name of this stream */
    char            type;	/* type  " */
    int             indprox;	/* index of associated Prox in aProx */
}               DescFid;

static DescFid        *aDescFid = (DescFid *) NULL;
static Boolean         replay = False;
static ASAPstatus_t    err;
static HashTable      *htKnownStream;
int                    global_argc;
char                 **global_argv;


/**********/
/* protos */

int             main(int argc, char **argv, char **env);
static void     getEnvParams(char **psfn, char **pdfn, char **prfn);
static void     parseParams(int *pargc, char ***pargv,
			                    char **psfn, char **pdfn, char **prfn, char **pss, Boolean * pmaxSpeed);
static void     usage(void);
static void     completeParams(int argc, char **argv, char **psfn, char **pdfn, char **prfn);
static void     startReplay(char *rfn, FILE ** pdf);
static void     startAlgo(int argc, char **argv, char *dfn, ASAPserv_t * pserv, ASAPconn_t * pconn);
static void     setEnvVars(char *dfn, Boolean play);
static int      getNewValueFromFile(FILE * df);
static Value   *getNewValueFF(FILE * df, Byte type);
static long     getWaitTimeFromFile(FILE * df);
static Value   *getNewSingleValueFF(FILE * df, int vtype);
static int      getNewValueFromAlgo(ASAPserv_t serv, ASAPconn_t conn);
static Value   *getNewValueFA(channel_t chan, MsgHeader mh);
static Value   *getNewSingleValueFA(channel_t chan, unsigned char vtype, int fid);
static void     checkAsapStat(ASAPstatus_t stat, char *errMsg);
static void     storeNewStreamNameFromAlgo(channel_t chan, MsgHeader mh);
static void     checkNewStream(char *name, Byte type);
static void     addNewName(char *name, Byte type, int ip);
static void     checkSameType(MsgHeader mh);
static void    *keyDescFid(DescFid * pd);
static void    *keyXwin(Xwin * pxw);




/**********/
/* bodies */


int             main(int argc, char **argv, char **env)
{
    Boolean         t = Ok;
    long            wt;		/* waiting time during replay */
    char           *sfn = NULL;	/* Script File Name */
    char           *dfn = NULL;	/* Dump File Name */
    char           *rfn = NULL;	/* Replay File Name */
    char           *sspeed = NULL;	/* Speed */
    ASAPserv_t      serv;
    ASAPconn_t      conn;
    FILE           *df;
    float           timeStep = 1.0;
    Boolean         maxSpeed = False;

    global_argc = argc;
    global_argv = argv;

    getEnvParams(&sfn, &dfn, &rfn);
    lIncludePath = lCreate(keySame, ordString, free, printString);
    lAdd(lIncludePath, (void *) ".");

    parseParams(&argc, &argv, &sfn, &dfn, &rfn, &sspeed, &maxSpeed);
    if (sspeed != NULL) {
	if (maxSpeed == True)
	    warn("Speed option -s discarded. MaxSpeed option -m used instead.");
	timeStep = 100.0 / atoi(sspeed);
    }
    completeParams(argc, argv, &sfn, &dfn, &rfn);
    setEnvVars(dfn, (sfn != NULL) ? True : False);
    htKnownStream = htCreate(HT_KNOWNSTREAM_SZ, keyDescFid, ordString, nullFunc, hashString, printString);
    htXwin = htCreate(HT_XWIN_SZ, keyXwin, ordInt, nullFunc, hashInt, printInt);
    if (replay == False) {
	startAlgo(argc, argv, dfn, &serv, &conn);
    }
    else {
	startReplay(rfn, &df);
    }
    compileFile(sfn);
    while (1) {
	if (replay == False) {
	    t = getNewValueFromAlgo(serv, conn);
	}
	else {
	    t = getNewValueFromFile(df);
	    wt = getWaitTimeFromFile(df);
	    if (!maxSpeed) {
		struct timeval  time;

		wt = timeStep * wt;
		time.tv_sec = wt / 1000;
		time.tv_usec = (wt % 1000) * 1000;
		select(0, NULL, NULL, NULL, &time);
	    }
	}
	if (t != Ok)
	    break;
	runProx();
	checkXEvents();
    }
    if (argv[0] != NULL)
	printf("\n\nAgat: End of %s\n", argv[0]);
    else if (rfn != NULL)
	printf("\n\nAgat: End of replay for %s\n", rfn);
    else
	printf("\n\nAgat: End of process\n");


    while (1) {
	checkXEvents();
	usleep(10000); /* 10ms */
    }
    return 0;
}

static void     getEnvParams(char **psfn, char **pdfn, char **prfn)
{
    char            tmp[MAX_CHAR_TMP];

    sprintf(tmp, "%s%s", AGAT_ENV, "script");
    *psfn = getenv(tmp);
}

static void     parseParams(int *pargc, char ***pargv, char **psfn, char **pdfn, char **prfn, char **pss, Boolean * pmaxSpeed)
{
    Boolean         agtParams = True;
    Boolean         fSet = False;
    char           *pathComp;

    (*pargc)--;			/* forget the name of the command */
    (*pargv)++;
    if (*pargc == 0)
	usage();
    while ((agtParams == True) && (*pargc != 0)) {
	if ((**pargv)[0] != '-') {
	    agtParams = False;
	}
	else {
	    switch ((**pargv)[1]) {
	    case 'f':{
		    if (fSet == True) {
			warn2(WM_OPTMULTDEF, "-f");
			warn2(WM_USING, *psfn);
		    }
		    else if (*psfn != NULL) {
			warn2(WM_ENVARSET, "-f");
			if (getStringParam(pargc, pargv, psfn) != Ok)
			    usage();
			warn2(WM_USING, *psfn);
		    }
		    else if (getStringParam(pargc, pargv, psfn) != Ok) {
			usage();
		    }
		    fSet = True;
		    break;
		}
	    case 'd':{
		    if (*pdfn != NULL) {
			warn2(WM_OPTMULTDEF, "-d");
			warn2(WM_USING, *pdfn);
		    }
		    else if (getStringParam(pargc, pargv, pdfn) != Ok) {
			usage();
		    }
		    break;
		}
	    case 'r':{
		    if (*prfn != NULL) {
			warn2(WM_OPTMULTDEF, "-r");
			warn2(WM_USING, *prfn);
		    }
		    else if (getStringParam(pargc, pargv, prfn) != Ok) {
			usage();
		    }
		    replay = True;
		    break;
		}
	    case 's':{
		    if (*pss != NULL) {
			warn2(WM_OPTMULTDEF, "-s");
			warn2(WM_USING, *pss);
		    }
		    else if (getStringParam(pargc, pargv, pss) != Ok) {
			usage();
		    }
		    break;
		}
	    case 'I':{
		    if (getStringParam(pargc, pargv, &pathComp) != Ok) {
			usage();
		    }
		    lAdd(lIncludePath, pathComp);
		    break;
		}
	    case 'm':{
		    *pmaxSpeed = True;
		    (*pargc)--;
		    (*pargv)++;
		    break;
		}
	    case 'w':{
		    silentWarn = True;
		    (*pargc)--;
		    (*pargv)++;
		    break;
		}
	    default:{
		    usage();
		    break;
		}
	    }
	}
    }
}


static void     usage(void)
{
    fprintf(stderr, "agat [-f scriptname] [-d dumpfilename] [-r replayfilename] [-s speed] [-m] [-w] [cmd] [parameters of cmd]\n");
    fprintf(stderr, "     scriptname    : name of the .agt file to use for animation\n");
    fprintf(stderr, "     dumpfilename  : name of the file to use for animation storage\n");
    fprintf(stderr, "     replayfilename: name of the file to use for animation replay\n");
    fprintf(stderr, "     speed         : an integer to use as speed ratio during replay\n                     (100 = normal speed)\n");
    fprintf(stderr, "     -m            : replay animation with maximum speed\n");
    fprintf(stderr, "     -w            : supress warnings\n");
    fprintf(stderr, "     cmd           : name of the executable command to animate\n");
    fprintf(stderr, "     parameters of : all the parameters of the cmd \n");
    exit(1);
}



static void     completeParams(int argc, char **argv, char **psfn, char **pdfn, char **prfn)
{
    if ((*prfn != NULL) && (argc != 0)) {
	warn2(WM_REPLFNUNUSED, argv[0]);
	free(*prfn);
	*prfn = NULL;
	replay = False;
    }
    if ((*prfn != NULL) && (*pdfn != NULL)) {
	warn2(WM_REPLFANDDMPFN, *prfn);
	free(*pdfn);
	*pdfn = NULL;
    }
    if ((*prfn == NULL) && (argc == 0)) {
	error(ER_NOANIM);
    }
    if ((*psfn == NULL) && (((argc != 0) && (*pdfn == NULL)) || (*prfn != NULL))) {
	error(ER_NOSFN);
    }
}


static void     startReplay(char *rfn, FILE ** pdf)
{
    *pdf = UCompressedFOpen(rfn, "r");
    if (*pdf == NULL) {
	error3(ER_CANTOPENFILE, rfn, " (the replay file)");
    }
}


static void     startAlgo(int argc, char **argv, char *dfn, ASAPserv_t * pserv, ASAPconn_t * pconn)
{

    initServ(pserv);
    pserv->type = SERV_LOCAL;
    pserv->cmd = (char *) argv;
    if ((err = ASAPconnect(*pserv, pconn)) != ASAP_SUCCESS) {
	error2("server not connected", "tag");
    }

}


static void     setEnvVars(char *dfn, Boolean play)
{
    char            tmp[MAX_CHAR_TMP];

    if (dfn != NULL) {
	sprintf(tmp, "%s%s=%s", AGAT_ENV, "dump", dfn);
	putenv(myStrDup(tmp));
    }
    else {
	sprintf(tmp, "%s%s=%s", AGAT_ENV, "dump", "");
	putenv(myStrDup(tmp));
    }

    if (play) {
	sprintf(tmp, "%s%s=%s", AGAT_ENV, "play", "ok");
	putenv(myStrDup(tmp));
    }
    else {
	sprintf(tmp, "%s%s=%s", AGAT_ENV, "play", "");
	putenv(myStrDup(tmp));
    }
    sprintf(tmp, "%s%s=%s", AGAT_ENV, "running", "ok");
    putenv(myStrDup(tmp));

}


static int      getNewValueFromFile(FILE * df)
{
    char            fn[FN_MAXLEN], *name;
    DescFid        *pd;
    Value          *pv;
    int             type;
    int             indProx, fs;

    fs = fscanf(df, "%s %d ", fn, &type);
    if ((fs == EOF) || (fs < 2))
	return Err;
    name = myStrDup(fn);
    /* Is it the first reference to this Stream */
    if ((pd = htSearchKey(htKnownStream, name)) == NULL) {
	checkNewStream(name, (Byte) type);
	pd = htSearchKey(htKnownStream, name);
    }
    pv = getNewValueFF(df, (Byte) type);
    if (pv != NULL) {
	if ((indProx = pd->indprox) >= 0) {
	    /* check if this stream is part of evaluation graph */
	    /* a stream can be send by algorithm and no be used */
	    giveValueToSuccs(aProx[indProx], pv);
	}
    }
    return Ok;
}

static Value   *getNewValueFF(FILE * df, Byte type)
{
    int             vtype, stype;

    vtype = EXTRACT_VTYPE(type);
    stype = EXTRACT_STYPE(type);
    switch (stype) {
    case STYP_SVAL:{
	    return getNewSingleValueFF(df, vtype);
	    break;
	}
    case STYP_TAB:{
	    warn("unimplemented type of data: STYP_TAB");
	    return NULL;
	    break;
	}
    default:{
	    warn2(ERS_UNKNOWN_STYP, charToBit(stype));
	    return NULL;
	    break;
	}
    }
}


static long     getWaitTimeFromFile(FILE * df)
{
    long            wt;

    if (fscanf(df, "%d", &wt) != 1)
	return 0;
    return wt;
}


static Value   *getNewSingleValueFF(FILE * df, int vtype)
{
    Value          *pv;

    long            i;
    float           f;
    double          d;


    pv = allocValue();
    switch (vtype) {
    case VTYP_CHAR:{
	    if (fscanf(df, "%d", &i) != 1)
		return NULL;
	    pv->v.c = (char) i;
	    pv->t = VT_CHAR;
	    return pv;
	    break;
	}
    case VTYP_SHORT:{
	    if (fscanf(df, "%d", &i) != 1)
		return NULL;
	    pv->v.s = (short) i;
	    pv->t = VT_SHORT;
	    return pv;
	    break;
	}
    case VTYP_INT:{
	    if (fscanf(df, "%d", &i) != 1)
		return NULL;
	    pv->v.i = (int) i;
	    pv->t = VT_INT;
	    return pv;
	    break;
	}
    case VTYP_LONG:{
	    if (fscanf(df, "%d", &i) != 1)
		return NULL;
	    pv->v.l = (long) i;
	    pv->t = VT_LONG;
	    return pv;
	    break;
	}
    case VTYP_UCHAR:{
	    if (fscanf(df, "%d", &i) != 1)
		return NULL;
	    pv->v.uc = (unsigned char) i;
	    pv->t = VT_UCHAR;
	    return pv;
	    break;
	}
    case VTYP_USHORT:{
	    if (fscanf(df, "%d", &i) != 1)
		return NULL;
	    pv->v.us = (unsigned short) i;
	    pv->t = VT_USHORT;
	    return pv;
	    break;
	}
    case VTYP_UINT:{
	    if (fscanf(df, "%d", &i) != 1)
		return NULL;
	    pv->v.ui = (unsigned int) i;
	    pv->t = VT_UINT;
	    return pv;
	    break;
	}
    case VTYP_ULONG:{
	    if (fscanf(df, "%d", &i) != 1)
		return NULL;
	    pv->v.ul = (unsigned long) i;
	    pv->t = VT_ULONG;
	    return pv;
	    break;
	}

    case VTYP_FLOAT:{
	    if (fscanf(df, "%f", &f) != 1)
		return NULL;
	    pv->v.f = (float) f;
	    pv->t = VT_FLOAT;
	    return pv;
	    break;
	}
    case VTYP_DOUBLE:{
	    if (fscanf(df, "%lf", &d) != 1)
		return NULL;
	    pv->v.d = (double) d;
	    pv->t = VT_DOUBLE;
	    return pv;
	    break;
	}
    default:{
	    warn2(ERS_UNKNOWN_VTYP, charToBit(vtype));
	    return NULL;
	}
    }
}




static int      getNewValueFromAlgo(ASAPserv_t serv, ASAPconn_t conn)
{
    ASAPstatus_t    stat;
    MsgHeader       mh;
    int             indProx, norb = 0;
    Value          *pv;
    channel_t       chan;

    chan = conn.normal;
    stat = ASAPgetBinary(chan, (char *) &mh, sizeof(MsgHeader), &norb);
    if ((stat == ASAP_BROKEN) || (stat == ASAP_EBADTYPE))
	return Err;
    checkAsapStat(stat, "get msg header: ");
    if (mh.fid == 0) {
	storeNewStreamNameFromAlgo(chan, mh);
	/* wait for a new Msg to get value */
	stat = ASAPgetBinary(chan, (char *) &mh, sizeof(MsgHeader), &norb);
	if ((stat == ASAP_BROKEN) || (stat == ASAP_EBADTYPE))
	    return Err;
	checkAsapStat(stat, "get msg header: ");
	if (mh.fid == 0)
	    error(" internal error: intricated stream declaration!?");
    }

    pv = getNewValueFA(chan, mh);
    if (pv != NULL) {
	if ((indProx = aDescFid[mh.fid].indprox) >= 0) {
	    /* check if this stream is part of evaluation graph */
	    /* a stream can be send by algorithm and no be used */
	    giveValueToSuccs(aProx[indProx], pv);
	}
    }
    return Ok;
}


static Value   *getNewValueFA(channel_t chan, MsgHeader mh)
{
    Byte            vtype;	/* for value type */
    Byte            stype;	/* for structure type */

    vtype = EXTRACT_VTYPE(mh.type);
    stype = EXTRACT_STYPE(mh.type);
    checkSameType(mh);
    switch (stype) {
    case STYP_SVAL:{
	    return getNewSingleValueFA(chan, vtype, mh.fid);
	    break;
	}
    case STYP_TAB:{
	    warn("unimplemented type of data: STYP_TAB");
	    return NULL;
	    break;
	}
    default:{
	    warn2(ERS_UNKNOWN_STYP, charToBit(stype));
	    return NULL;
	    break;
	}
    }
}


static Value   *getNewSingleValueFA(channel_t chan, unsigned char vtype, int fid)
{
    int             norb;
    ASAPstatus_t    stat;
    Value          *pv;

    char            c;
    short           s;
    int             i;
    long            l;
    unsigned char   uc;
    unsigned short  us;
    unsigned int    ui;
    unsigned long   ul;
    float           f;
    double          d;


    pv = allocValue();
    switch (vtype) {
    case VTYP_CHAR:{
	    stat = ASAPgetBinary(chan, (char *) &c, sizeof(char), &norb);
	    checkAsapStat(stat, "get Value char: ");
	    pv->v.c = c;
	    pv->t = VT_CHAR;
	    return pv;
	    break;
	}
    case VTYP_SHORT:{
	    stat = ASAPgetBinary(chan, (char *) &s, sizeof(short), &norb);
	    checkAsapStat(stat, "get Value short: ");
	    pv->v.s = s;
	    pv->t = VT_SHORT;
	    return pv;
	    break;
	}
    case VTYP_INT:{
	    stat = ASAPgetBinary(chan, (char *) &i, sizeof(int), &norb);
	    checkAsapStat(stat, "get Value int: ");

	    pv->v.i = i;
	    pv->t = VT_INT;
	    return pv;
	    break;
	}
    case VTYP_LONG:{
	    stat = ASAPgetBinary(chan, (char *) &l, sizeof(long), &norb);
	    checkAsapStat(stat, "get Value long: ");
	    pv->v.l = l;
	    pv->t = VT_LONG;
	    return pv;
	    break;
	}
    case VTYP_UCHAR:{
	    stat = ASAPgetBinary(chan, (char *) &uc, sizeof(unsigned char), &norb);
	    checkAsapStat(stat, "get Value unsigned char: ");
	    pv->v.uc = uc;
	    pv->t = VT_UCHAR;
	    return pv;
	    break;
	}
    case VTYP_USHORT:{
	    stat = ASAPgetBinary(chan, (char *) &us, sizeof(unsigned short), &norb);
	    checkAsapStat(stat, "get Value unsigned short: ");
	    pv->v.us = us;
	    pv->t = VT_USHORT;
	    return pv;
	    break;
	}
    case VTYP_UINT:{
	    stat = ASAPgetBinary(chan, (char *) &ui, sizeof(unsigned int), &norb);
	    checkAsapStat(stat, "get Value unsigned int: ");
	    pv->v.ui = ui;
	    pv->t = VT_UINT;
	    return pv;
	    break;
	}
    case VTYP_ULONG:{
	    stat = ASAPgetBinary(chan, (char *) &ul, sizeof(unsigned long), &norb);
	    checkAsapStat(stat, "get Value unsigned long: ");
	    pv->v.ul = ul;
	    pv->t = VT_ULONG;
	    return pv;
	    break;
	}

    case VTYP_FLOAT:{
	    stat = ASAPgetBinary(chan, (char *) &f, sizeof(float), &norb);
	    checkAsapStat(stat, "get Value float: ");
	    pv->v.f = f;
	    pv->t = VT_FLOAT;
	    return pv;
	    break;
	}
    case VTYP_DOUBLE:{
	    stat = ASAPgetBinary(chan, (char *) &d, sizeof(double), &norb);
	    checkAsapStat(stat, "get Value double: ");
	    pv->v.d = d;
	    pv->t = VT_DOUBLE;
	    return pv;
	    break;
	}
    default:{
	    warn2(ERS_UNKNOWN_VTYP, charToBit(vtype));
	    return NULL;
	}
    }
}


static void     checkAsapStat(ASAPstatus_t stat, char *errMsg)
{
    if ((stat != ASAP_SUCCESS) && (stat != ASAP_FINISH))
	error3("serv: ", errMsg, ASAPerrorText(stat));
}


static void     storeNewStreamNameFromAlgo(channel_t chan, MsgHeader mh)
{
    ASAPstatus_t    stat;
    SDescStream     sf;
    char           *name;
    int             norb;

    stat = ASAPgetBinary(chan, (char *) &sf, sizeof(SDescStream), &norb);
    checkAsapStat(stat, "get Stream descriptor: ");

    name = (char *) malloc(sizeof(char) * sf.nameLen);
    stat = ASAPgetBinary(chan, name, sf.nameLen, &norb);
    checkAsapStat(stat, "get Stream name: ");
    checkNewStream(name, mh.type);
}

static void     checkNewStream(char *name, Byte type)
{
    ExtStream      *pes;
    int             indprox;

    /* Is this Stream a good one?                        */
    /* (used in the .agt file which is beeing processed) */
    if ((pes = htSearchKey(htExtStream, name)) == NULL) {
	char            tmp[MAX_CHAR_TMP];

	sprintf(tmp, "incoming data on stream %s will be ignored (unknown stream)", name);
	evalWarning(NULL, 0, tmp);
	indprox = -1;
    }
    else
	indprox = pes->indProx;
    addNewName(name, type, indprox);
}


static void     addNewName(char *name, Byte type, int ip)
{
    static int      aDescFidSz = 0;
    static int      nbDescFid = 0;

    if (nbDescFid >= aDescFidSz) {
	if (aDescFid == (DescFid *) NULL) {
	    aDescFidSz += RM_CHKZ_DESCFID;
	    aDescFid = (DescFid *) malloc(sizeof(DescFid) * aDescFidSz);
	    nbDescFid = 1;
	    /* the first fid is 1 to distinguish no-fid==0 and first-fid==1 */
	}
	else {
	    aDescFidSz += RM_CHKZ_DESCFID;
	    aDescFid = (DescFid *) realloc(aDescFid, sizeof(DescFid) * aDescFidSz);
	}
    }
    aDescFid[nbDescFid].name = name;
    aDescFid[nbDescFid].type = type;
    aDescFid[nbDescFid].indprox = ip;
    htInsert(htKnownStream, &aDescFid[nbDescFid]);
    nbDescFid++;
}


static void     checkSameType(MsgHeader mh)
{
    if ((int) aDescFid[mh.fid].type != (int) mh.type)
	warn2(WN_CHANGEDMETYPE, aDescFid[mh.fid].name);
}


static void    *keyDescFid(DescFid * pd)
{
    return pd->name;
}

static void    *keyXwin(Xwin * pxw)
{
    return &pxw->win;
}
